import pathlib

from seaquest import experiment, monitor, uploader
from seaquest.utils import validate

from test_vars import *

def cleanup():
    args = validate._parse_args(COMPLETE_ARGS.copy())

    job_name = "{prefix}-{model}-{fun}-{data_file}-job{suffix}-{idx}".format(prefix=prefix, model=args["model_name"], fun=args["model_fun"], idx=0, data_file=data_file, suffix="")
    job_name = job_name.lower()
    monitor._delete_job(api_instance_batch, namespace, job_name)
    uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)


def test_upload_files():
    try:
        args = validate._parse_args(COMPLETE_ARGS.copy())

        experiment._upload_files(args)
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
        assert False, "[experiment] upload_files failed with error: {e}".format(e=e)


def test_launch_jobs():
    try:
        args = validate._parse_args(COMPLETE_ARGS.copy())

        experiment._launch_jobs(args, pvc_name_prefix)
        cleanup()
    except Exception as e:
        cleanup()
        assert False, "[experiment] upload_files failed with error: {e}".format(e=e)


def test_main_func():
    try:
        experiment.main(COMPLETE_ARGS.copy())
        cleanup()
    except Exception as e:
        cleanup()
        assert False, "[experiment] upload_files failed with error: {e}".format(e=e)


from importlib import import_module
import os
import pathlib
import pkgutil
import sys


from seaquest.utils.loggus import init_logger
from seaquest.utils.validate import _parse_runner_args

logger = init_logger(__name__ if __name__ != "__main__" else pathlib.Path(__file__).stem, level="debug")


def _load_model_dir(model_dir):
    """Load model directory

    Parameters
    ----------
    model_dir: str
        The name of the model directory
    Returns
    -------
    namespace: str
        The name of the namespace where the model can be loaded from
    """
    logger.info("Attempting to load model directory ...")
    try:
        namespace = import_module(model_dir)
    except ModuleNotFoundError as e:
        raise ValueError(f"Model directory '{model_dir}' or one of its imports not found or is not a module!") from e

    logger.info("Model directory successfully loaded!")
    return namespace


def _load_model_class(namespace, model_name):
    """Loads the model class that instantiates a model object

    Parameters
    ----------
    namespace: str
        The namespace that holds the model
    model_name: str
        The name of the model
    Returns
    -------
    model: class
        The model class that instantiates a model object
    """

    logger.info("Attempting to load model class ...")
    subpackages = [name for _, name, _ in list(pkgutil.iter_modules(
        namespace.__path__,
        namespace.__name__ + "."
    ))]
    
    # find the model and load it
    found_model = False
    for subp in subpackages:
        try:
            imp_subp = import_module(subp)
            model = getattr(imp_subp, model_name)
            found_model = True
        except ModuleNotFoundError as e:
            pass
        except AttributeError as e:
            pass

    if found_model is False:
        raise ModuleNotFoundError(f"Module '{model_name}' not found in directory {namespace.__name__}")
    logger.info(f"Successfully imported model '{model_name}' from directory '{namespace.__name__}'")

    return model


def _load_model_function(model, fun):
    callable = getattr(model, fun, None)
    if callable is None:
        raise ValueError(f"Function '{fun}' not found in model '{model.__name__}'")

    return callable


def main(args: dict) -> None:
    """Dynamically imports and runs a specified function from a model class.
    
    Parameters:
    -----------
        model_name: str 
            Name of the model to import.
        model_dir: str 
            Directory where the model is located.
        data_dir: str 
            Directory containing the data for training or inference.
        output_dir: str 
            Directory where the output should be stored.
        fun:
          str Function to execute from the model class ('train' or 'infer').
    
    Returns:
    --------
        None
    
    Raises:
    -------
        ValueError: If the model or function cannot be found.
    """

    output_dir = pathlib.Path(args["output_dir"]) # .resolve()
    model_dir = pathlib.Path(pathlib.Path.cwd() / args["md_dir"])
    data_file = pathlib.Path(args["data_file"])
    sys.path.insert(0, str(model_dir.parent)) # needs str

    # TODO: refactor this?
    namespace = _load_model_dir(model_dir=model_dir.name)
    model_class = _load_model_class(namespace=namespace, model_name=args["model_name"])

    # change dir to model_dir so that all relative imports of the model work
    os.chdir(os.path.dirname(namespace.__file__))

    if "model_keyword_args" in args:
        model = model_class(output_dir=output_dir, data_file=data_file, **args["model_keyword_args"])
    else:
        model = model_class(output_dir=output_dir, data_file=data_file)

    # run the experiments
    logger.info("Starting function '{fun}' from model '{mn}'".format(fun=args["model_fun"], mn=args["model_name"]))
    
    if args["model_fun"] == "train":
        model.train()

    elif args["model_fun"] == "infer":
        model.infer()

    else:
        raise ValueError("Provided function should be either test or infer (was {fun})".format(fun=args["model_fun"]))

    logger.info("Function '{fun}' from model '{mn}' completed successfully".format(fun=args["model_fun"], mn=args["model_name"]))


if __name__ == "__main__":
    args = _parse_runner_args(sys.argv)
    main(args)


VALIDATION_SCHEMA = {
    "config_file": {
        "type": "string", 
        "required": True
    },
    "md_dir": {
        "type": "string", 
        "required": True
    },
    "data_file": {
        "type": "string",
        "required": True
    },
    "model_name": {
        "type": "string", 
        "required": True
    },
    "model_fun": {
        "type": "string", 
        "required": True
    },
    "prefix": {
        "type": "string", 
        "required": True,
        "default": "",
    },
    "suffix": {
        "type": "string",
        "default": ""
    },
    "log_file": {
        "type": "string"
    },
    "job_spec": {
        "type": "dict",
        "schema": {
            "resources": {
                "type": "dict", 
                "schema": {
                    "limits": {
                        "type": "dict",
                        "schema": {
                            "memory": {"type": "string", "required": True},
                            "cpu": {"type": "integer", "required": True},
                            "nvidia.com/gpu": {"type": "integer"},
                        }
                    },
                    "requests": {
                        "type": "dict",
                        "schema": {
                            "memory": {"type": "string", "required": True},
                            "cpu": {"type": "integer", "required": True},
                            "nvidia.com/gpu": {"type": "integer"},
                        }
                    },
                }
            },
            "graphics-card": {
                "type": "string",
            },
        }
    },
    "kube_env": {
        "type": "dict",
        "schema": {
            "namespace": {
                "type": "string", 
                "required": True
            },
        }
    },
    "pvc_params": {
        "type": "dict",
        "schema": {
            "pvc-name": {
                "type": "string"
            }
        }
    },
    "model_keyword_args": {
        "type": "dict",
    }
}


from argparse import ArgumentParser
from cerberus import Validator
import collections.abc
import pathlib
from yaml import safe_load

from .vschema import VALIDATION_SCHEMA


def _update(d: dict, u: dict):
    """merge two dictionaries, keeping the values of second one in case of key collision
    
     Parameters
     ----------
     d: dict
          First dictionary
     u: dict
          Second dictionary

     Returns
     -------
     d: dict
          Merged dictionary
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = _update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def parse_and_validate_args(argv) -> dict:
     """parse and validate the command line arguments

     Parameters
     ----------
     argv: list
          List of command line arguments

     Returns
     -------
     args: dict
          Dictionary of parsed arguments
     """
     args = _parse_args(argv)
     _validate(args)
     return args


def _validate(args: dict) -> bool:
     """validate the parsed arguments
     Check if all the mandatory arguments are provided
     
     Parameters
     ----------
     args: dict
          Dictionary of parsed arguments
     Returns
     -------
     bool
          True if all the mandatory arguments are provided, else raise ValueError
     """
     
     validator = Validator(VALIDATION_SCHEMA)
     validator.validate(args)

     if validator.errors != {}:
          raise KeyError("Not all required arguments have been provided or provided arguments are incorrect: {err}".format(err=validator.errors))

     return True


def _parse_args(argv) -> dict:
     """parse the command line arguments

     Parameters
     ----------
     argv: list
          List of command line arguments
     Returns
     -------
     args: dict
          Dictionary of parsed arguments
     """

     parser = ArgumentParser()

     # required args
     parser.add_argument("-cf", "--config-file", action="store", dest="config_file", required=True, default="./configs/default_job_cfg.yaml", type=str, help="Path to config file")

     # optional args
     parser.add_argument("-lf", "--log-file", action="store", dest="log_file", default="seaquest_logs.logs", required=False, type=str, help="Logging file path")

     parser.add_argument("-md", "--model-data-dir", action="store", dest="md_dir", required=False, type=str, help="Path to directory containing model and data")
     parser.add_argument("-df", "-data-file", action="store", dest="data_file", required=False, type=str, help="Path to file containing the data for the model")
     parser.add_argument("-mn", "--model-name", action="store", dest="model_name", required=False, type=str, help="Name of the model")
     parser.add_argument("-f", "--model-function", action="store", dest="model_fun", required=False, type=str, help="Name of the model function to run", choices=['train', 'infer'])
     parser.add_argument("-p", "--prefix", action="store", dest="prefix", required=False, type=str, help="Prefix to use for naming jobs and other k8s objects")
     parser.add_argument("-s", "--suffix", action="store", dest="suffix", required=False, type=str, help="Suffix to use for naming jobs and other k8s objects")

     cmd_args, _ = parser.parse_known_args(argv)
     config_args = _parse_config_file(pathlib.Path(cmd_args.config_file).resolve())

     return _update(vars(cmd_args), config_args)


def _parse_runner_args(argv) -> dict:
     """parse the command line arguments for the runner

     Parameters
     ----------
     argv: list
          List of command line arguments
     Returns
     -------
     args: dict
          Dictionary of parsed arguments
     """

     parser = ArgumentParser()

     parser.add_argument("-cf", "--config-file", action="store", dest="config_file", required=True, type=str, help="Path to config file")
     
     # parser.add_argument("-md", "--model-data-dir", action="store", dest="model_dir", required=False, type=str, help="Path to directory containing model and data")
     # parser.add_argument("-df", "-data-file", action="store", dest="data_file", required=False, type=str, help="Path to file containing the data for the model")
     # parser.add_argument("-mn", "--model-name", action="store", dest="model_name", required=False, type=str, help="Name of the model")
     # parser.add_argument("-f", "--model-function", action="store", dest="model_fun", required=False, type=str, help="Name of the model function to run", choices=['train', 'infer'])
     parser.add_argument("-od", "--output-dir", action="store", dest="output_dir", required=True, type=str, help="output_dir")

     cmd_args, _ = parser.parse_known_args(argv)
     config_args = _parse_config_file(pathlib.Path(cmd_args.config_file).resolve())

     return _update(vars(cmd_args), config_args)


def _parse_monitor_args(argv) -> dict:
     """parse the command line arguments for the monitor

     Parameters
     ----------
     argv: list
          List of command line arguments
     Returns
     -------
     args: dict
          Dictionary of parsed arguments
     """

     parser = ArgumentParser()

     parser.add_argument("-cf", "--config-file", action="store", dest="config_file", required=False, type=str, help="Path to config file")
     parser.add_argument("-d", "--dest-dir", action="store", dest="dest_dir", required=True, type=str, help="Path to where the output should be saved")
     
     cmd_args, _ = parser.parse_known_args(argv)

     return vars(cmd_args)


def _parse_config_file(file_path: pathlib.Path) -> dict:
     """parse a yaml config file

     Parameters
     ----------
     file_path: str
          Path to the config file

     Returns
     -------
     config: dict
          Dictionary of parsed config file
     """
     with open(file_path, 'r') as f:
          config = safe_load(f)

     return config


from abc import ABC, abstractmethod
import os
import pathlib


class OutputContext:
    """Class that handles the context switch for saving output files on the PVC"""

    def __init__(self, output_dir: pathlib.Path):
        self.output_dir = output_dir

    def __enter__(self):
        """"""
        self.curr_wd = os.getcwd()
        os.chdir(self.output_dir) # .resolve())

        return self

    def __exit__(self, exc_type, exc_value, traceback):
       """"""
       os.chdir(self.curr_wd)


class NautPipelineModel(ABC):
    """Base class to be inherited by all models to ensure minimum required constructor arguments and methods"""
    def __init__(self, output_dir: pathlib.Path, data_file: pathlib.Path):
        super().__init__()

        self.output_dir = output_dir
        self.data_file = data_file

    @abstractmethod
    def train(self):
        pass

    @abstractmethod
    def infer(self):
        pass


import itertools
import pathlib

from kubernetes import client
from yaml import safe_dump

from seaquest.utils.loggus import init_logger

logger = init_logger(__name__ if __name__ != "__main__" else pathlib.Path(__file__).stem, level="debug")


def _prepare_afinity(job_config: dict) -> client.V1Affinity:
    """create the affinity object for the k8s job

    Parameters
    ----------
    job_config: dict
        Dictionary of parsed job configuration arguments
    Returns
    -------
    affinity: client.V1Affinity
        kubernetes affinity object
    """
    return client.V1Affinity( # TODO: make this configurable
                node_affinity=client.V1NodeAffinity(
                    required_during_scheduling_ignored_during_execution=client.V1NodeSelector(
                        node_selector_terms=[client.V1NodeSelectorTerm(
                            match_expressions=[client.V1NodeSelectorRequirement(
                                key="nvidia.com/gpu.product",
                                operator="In",
                                values=[job_config["graphics-card"]],
                            )]
                        )]
                    )
                )
            )


def _prepare_containers(config: dict, job_name: str, pvc: str, arguments: dict) -> list[client.V1Container]:
    """create the container objects for the k8s job

    Parameters
    ----------
    config: dict
        Dictionary of parsed job configuration arguments
    job_name: str
        The name of the job
    arguments: dict
        A list of arguments to pass to the runner module that will run inside the job
    Returns
    -------
    containers:  list[client.V1Container]
        list of kubernetes container objects
    """
    output_dir_name = arguments["-od"]
    requirements_path = str(pathlib.PurePosixPath(arguments["-md"]).joinpath("requirements.txt"))

    arguments["-od"] = str(pathlib.PurePosixPath("/").joinpath(pvc).joinpath(output_dir_name))
    list_arguments = " ".join(itertools.chain.from_iterable(zip(list(arguments.keys()), list(arguments.values()))))
    
    return [client.V1Container(
                name="program",
                image="iancuonescu/seaquest:latest",
                #image="alpine:latest",
                #image="huggingface/transformers-pytorch-gpu:4.41.2", # TODO: make this configurable
                volume_mounts=[client.V1VolumeMount(
                    name=pvc,
                    mount_path="/{pvc}".format(pvc=pvc)
                )],
                resources=client.V1ResourceRequirements(
                    limits=config["limits"],
                    requests=config["requests"]
                ),
                command=[
                    "/bin/sh",
                    "-c"
                    # 'tail', '-f', '/dev/null'
                ],
                args=[
                    "cd {pvc} && \
                    mkdir {copy_dir} -p && \
                    cp -R {source} {dest} && \
                    rm -R {source} && \
                    mkdir {output_dir} -p && \
                    cd {copy_dir} && \
                    python -m pip install -r {reqs} --no-build-isolation &&\
                    python -m seaquest.runner -cf /{pvc}/{copy_dir}/{source}/runner.yaml {arguments}".format(pvc=pvc, copy_dir=job_name, \
                                                                source=arguments["-md"], dest=job_name, \
                                                                output_dir=output_dir_name, \
                                                                reqs=requirements_path, \
                                                                arguments=list_arguments), # TODO: make this prettier (perhaps a shell script)
                ],
            )]


def _prepare_volumes(pvc: str) -> list[client.V1Volume]:
    """create the volume objects for the k8s job

    Parameters
    ----------
    pvc: str
        Name of the PVC
    Returns
    -------
    volumes: list[client.V1Volume]
        List of kubernetes volume objects to mount to the job
    """
    return [client.V1Volume(
                name=pvc,
                persistent_volume_claim=client.V1PersistentVolumeClaimVolumeSource(pvc)
            )]


def _prepare_job_spec(job_config: dict, job_name: str, pvc: str, arguments: dict) -> client.V1JobSpec:
    """create the specification for the k8s job

    Parameters
    ----------
    job_config: dict
        Dictionary of job configuration parameters
    job_name: str
        Name of the job
    pvc: str
        Name of the PVC
    arguments: dict
        List of arguments to pass to the runner module that runs inside the job
    Returns
    -------
    status: client.V1JobStatus
        Kubernetes job spec object
    """
    return client.V1JobSpec(
        backoff_limit=0, # TODO: make this configurable?
        # ttl_seconds_after_finished = 0, : # TODO: make this configurable
        template=client.V1PodTemplateSpec(
            spec=client.V1PodSpec(
                affinity=_prepare_afinity(job_config) if "graphics-card" in job_config else None,
                containers=_prepare_containers(job_config["resources"], job_name, pvc, arguments),
                volumes=_prepare_volumes(pvc),
                restart_policy="Never"
            )
        )
    )


def _launch_job(api_instance: client.BatchV1Api, namespace: str, job_name:str, job_config: dict, pvc: str, arguments: dict) -> None:
    """launches a kubernetes job based on the provided configuration 

    Parameters
    ----------
    api_instance: kubernetes client
        Kubernetes client
    namespace: str
        Kubernetes namespace
    job_name: str
        Name of the job
    job_config: dict
        Dictionary of job configuration parameters
    pvc: str
        Name of the PVC
    arguments: dict
        List of arguments to pass to the runner module that runs inside the job
    Returns
    -------
    None
    """
    logger.info("Launching job {job} ...".format(job=job_name))
    
    body = client.V1Job(
        api_version="batch/v1",
        kind="Job",
        metadata=client.V1ObjectMeta(name=job_name, namespace=namespace),
        spec=_prepare_job_spec(job_config, job_name, pvc, arguments),
        status=client.V1JobStatus(), # None?
    )

    api_instance.create_namespaced_job(namespace, body)
    logger.info("Succesfully launched job {job}!".format(job=job_name))


def create_jobs(api_instance: client.BatchV1Api, num_jobs: int, namespace: str, job_config: dict, prefix: str, suffix: str, model_name:str, model_fun: str, pvc: str, model_dir: str, data_file: str) -> list:
    """Launches 'num_jobs' kubernetes jobs based on the specified configuration using a unique naming assignment formed of the model name and the data file 

    Parameters
    ----------
    api_instance: kubernetes client
        Kubernetes client
    num_jobs: int
        Number of jobs to launch
    namespace: str
        Kubernetes namespace
    job_config: dict
        Dictionary of job configuration parameters
    prefix: str
        Prefix for the job name
    suffix: str
        Suffix for the job name
    model_name: str
        Name of the model to experiment with
    model_fun: str
        Name of the function called (either 'train' or 'infer')
    pvc: str
        Name of the PVC
    model_dir: str
        Name of the model directory
    data_file: str
        Name of the data file
    Returns
    -------
    all_created_jobs: list
        A list of names of the succesfully launched jobs
    """
    all_created_jobs = []
    for idx in range(num_jobs):
        job_name = "{prefix}-{model}-{fun}-{data_file}-job{suffix}-{idx}".format(prefix=prefix, model=model_name, fun=model_fun, idx=idx, data_file=data_file, suffix="" if suffix is None else "-{s}".format(s=suffix))
        job_name = job_name.lower() # kube convention

        arguments = {
            "-od": "{job_name}_output".format(job_name=job_name),
            "-md": model_dir
        }

        try:
            _launch_job(api_instance, namespace, job_name, job_config, pvc, arguments)
        except Exception as e:
            logger.critical("Failed to launch job number {num} due to error {err}".format(num=idx, err=e))
            raise e

        all_created_jobs.append(job_name)

    return all_created_jobs


import os

from seaquest.model_class import NautPipelineModel, OutputContext

class ExampleModel(NautPipelineModel):
    def __init__(self, output_dir, data_file, test_arg):
        super().__init__(output_dir, data_file)

        self.test_arg = test_arg

    def infer(self):
        with open('./test-weights.txt', 'r') as f:
            print("Loading weights from test_weights.txt")

        with OutputContext(self.output_dir):
            with open("delete_this.txt", "w") as f:
                f.write("Hello, this is a test!\n")
                f.write(self.test_arg)

            print("we are inside")
            print(os.getcwd())
        
        print("now we are back to")
        print(os.getcwd())

        print("this is the data dir")
        print(self.data_file)

    def train(self):
        with open('./test-weights.txt', 'r') as f:
            print("Loading weights from test_weights.txt")


from time import sleep

from seaquest import launcher, uploader
from seaquest.monitor import _delete_job
from seaquest.utils import validate
from test_vars import *


def test_launch_job_fail():
    args = validate._parse_args(COMPLETE_ARGS.copy())
    try:
        launcher._launch_job(api_instance_batch, "hocus-pocus-namespace", test_job_name_prefix, args["job_spec"], pvc_name_prefix, job_arguments)
    except Exception as e:
        pass


def test_launch_job():
    args = validate._parse_args(COMPLETE_ARGS.copy())
    args["output_dir"] = job_output_dir # workaround
    try:
        # uploader.upload_files_to_pvc(namespace, prefix, pvc_name, [ \
        #         [test_model_dir_path, None],
        #         [config_file_path, test_model_dir_path.name]
        #     ])
        
        launcher._launch_job(api_instance_batch, namespace, test_job_name_prefix, args["job_spec"], pvc_name_prefix, job_arguments)
        _delete_job(api_instance_batch, namespace, test_job_name_prefix)
        
        # uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
        # sleep(5) # for deletion changes to take effect
    except Exception as e:
        # uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
        _delete_job(api_instance_batch, namespace, test_job_name_prefix)
        assert False, "Job launch failed with error {e}".format(e=e)


def test_create_jobs():
    args = validate._parse_args(COMPLETE_ARGS.copy())
    args["output_dir"] = job_output_dir # workaround
    try:
        uploader.upload_files_to_pvc(namespace, prefix, pvc_name, [ \
                [test_model_dir_path, None],
                [config_file_path, test_model_dir_path.name]
            ])
        launcher.create_jobs(api_instance_batch, 1, namespace, args["job_spec"], prefix, None, args["model_name"], args["model_fun"], pvc_name_prefix, args["md_dir"], data_file)

        job_name = "{prefix}-{model}-{fun}-{data_file}-job{suffix}-{idx}".format(prefix=prefix, model=args["model_name"], fun=args["model_fun"], idx=0, data_file=data_file, suffix="")
        job_name = job_name.lower()
        _delete_job(api_instance_batch, namespace, job_name)
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)

        job_name = "{prefix}-{model}-{fun}-{data_file}-job{suffix}-{idx}".format(prefix=prefix, model=args["model_name"], fun=args["model_fun"], idx=0, data_file=data_file, suffix="")
        job_name = job_name.lower()
        _delete_job(api_instance_batch, namespace, job_name)
        assert False, "Job launch failed with error {e}".format(e=e)
    


import pathlib

from seaquest import uploader
from seaquest.utils import pod

from test_vars import *


def test_check_pvc_exists_false():
    try:
        # in case there is a pvc not deleted by another test
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
    except:
        pass
    try:
        ret = uploader._check_pvc_exists(api_instance_core, namespace, "hocus-pocus")
        assert ret == False, "PVC check failed"
    except Exception as e:
        assert False, f"PVC check failed with exception: {e}"


def test_create_pvc():
    try:
        uploader._create_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        assert False, f"PVC creation failed with exception: {e}"

    # test existing pvc
    try:
        uploader._create_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        assert False, f"PVC creation failed with exception: {e}"


def test_check_pvc_exists_true():
    try:
        ret = uploader._check_pvc_exists(api_instance_core, namespace, pvc_name_prefix)
        assert ret == True, "PVC check failed"
    except Exception as e:
        assert False, f"PVC check failed with exception: {e}"    


def test_copy_files_to_pod_pass():
    pod_name_tmp = None
    try:
        pod_name_tmp = pod.make_pod_name_unique(api_instance_core, namespace, pod_name)
        pod._launch_pod(api_instance_core, namespace, pod_name_tmp, pvc_name_prefix)
        pod._wait_for_running_state(api_instance_core, namespace, pod_name_tmp, pod_wait_timeout)
        
        uploader._copy_files_to_pod(api_instance_core, namespace, pod_name_tmp, test_model_dir_path, pathlib.PurePosixPath(pvc_name_prefix))
        
        # TODO: verify files are copied to pvc

        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
    except Exception as e:
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
        assert False, f"Copying files failed with exception: {e}"


def test_copy_files_to_pod_fail():
    pod_name_tmp = None
    try:
        pod_name_tmp = pod.make_pod_name_unique(api_instance_core, namespace, pod_name)
        uploader._create_pvc(api_instance_core, namespace, pvc_name_prefix)
        
        pod._launch_pod(api_instance_core, namespace, pod_name_tmp, pvc_name_prefix)
        pod._wait_for_running_state(api_instance_core, namespace, pod_name_tmp, pod_wait_timeout)
        
        # Copying from a non-existing path should fail
        uploader._copy_files_to_pod(api_instance_core, namespace, pod_name_tmp, pathlib.WindowsPath("C:/", "hocus-pocus"), pathlib.PurePosixPath(pvc_name_prefix))
        
        # TODO: verify files are copied to pvc

        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
    except Exception as e:
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
        assert True, f"Copying files failed with exception: {e}"


def test_upload_files_to_pvc():
    try:
        uploader.upload_files_to_pvc(namespace, prefix, pvc_name, [ \
                [test_model_dir_path, None],
                [config_file_path, test_model_dir_path.name]
            ])
    except Exception as e:
        assert False, f"Uploading files to pvc failed with exception: {e}"


def test_clean_up():
    try:
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        assert False, "Cleanup failed with exception: {e}".format(e=e)


import os
import pathlib
from time import sleep

from kubernetes import client, config

from seaquest import monitor, launcher, uploader
from seaquest.utils import validate, pod, loggus
from test_vars import *

logger = loggus.init_logger("test_monitor")

def test_delete_job_pass():
    try:
        args = validate.parse_and_validate_args(COMPLETE_ARGS.copy())
        launcher._launch_job(api_instance_batch, namespace, test_job_name_prefix, args["job_spec"], pvc_name_prefix, job_arguments)
        monitor._delete_job(api_instance_batch, namespace, test_job_name_prefix)
        sleep(5)
    except Exception as e:
        assert False, "Failed to delete job with err {e}".format(e=e)


def test_delete_job_fail():
    try:
        args = validate.parse_and_validate_args(COMPLETE_ARGS.copy())
        launcher._launch_job(api_instance_batch, namespace, test_job_name_prefix, args["job_spec"], pvc_name_prefix, job_arguments)
        monitor._delete_job(api_instance_batch, namespace, "hocus-pocus")

        # just in case
        monitor._delete_job(api_instance_batch, namespace, test_job_name_prefix)
    except Exception as e:
        monitor._delete_job(api_instance_batch, namespace, test_job_name_prefix)
        assert True, "Failed to delete job with err {e}".format(e=e)


def test_copy_files_from_pod():
    try:
        uploader.upload_files_to_pvc(namespace, prefix, pvc_name, files_path)
        pod._launch_pod(api_instance_core, namespace, pod_name_prefix, pvc_name_prefix)
        pod._wait_for_running_state(api_instance_core, namespace, pod_name_prefix, pod_wait_timeout)

        output_path = pathlib.Path.cwd().resolve().joinpath("tests//unit//outputs")
        output_path.mkdir(exist_ok=True)
        monitor._copy_files_from_pod(api_instance_core, pod_name_prefix, namespace, pathlib.Path("{p}/test_model_dir".format(p=pvc_name_prefix)), output_path)

        pod._delete_pod(api_instance_core, namespace, pod_name_prefix)
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        pod._delete_pod(api_instance_core, namespace, pod_name_prefix)
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
        sleep(5)
        assert False, "Test copy files from pod failed with error {e}".format(e=e)


def test_copy_files_from_pod_fail():
    try:
        uploader.upload_files_to_pvc(namespace, prefix, pvc_name, files_path)
        pod._launch_pod(api_instance_core, namespace, pod_name_prefix, pvc_name_prefix)
        pod._wait_for_running_state(api_instance_core, namespace, pod_name_prefix, pod_wait_timeout)

        output_path = pathlib.Path.cwd().resolve().joinpath("tests//unit/outputs")
        output_path.mkdir(exist_ok=True)
        monitor._copy_files_from_pod(api_instance_core, pod_name_prefix, namespace, pathlib.Path("{p}/hocus-pocus".format(p=pvc_name_prefix)), output_path)

        pod._delete_pod(api_instance_core, namespace, pod_name_prefix)
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        pod._delete_pod(api_instance_core, namespace, pod_name_prefix)
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
        sleep(5)
        assert True


def test_monitor_jobs():
    try:
        uploader._create_pvc(api_instance_core, namespace, pvc_name_prefix)

        args = validate.parse_and_validate_args(COMPLETE_ARGS.copy())
        launcher._launch_job(api_instance_batch, namespace, test_job_name_prefix, args["job_spec"], pvc_name_prefix, job_arguments)


        monitor._monitor_jobs(api_instance_batch, api_instance_core, namespace, [test_job_name_prefix], pvc_name_prefix, pathlib.WindowsPath("C:/", "Users", "Iancu", "Desktop"), prefix) 
        
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        monitor._delete_job(api_instance_batch, namespace, test_job_name_prefix)
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
        assert False, "Test monitor jobs failed with error {e}".format(e=e)


import logging


def init_logger(name: str, file_path: str = './seaquest.logs', level = logging.DEBUG) -> logging.Logger:
    """create an object that handles the logging

    Parameters
    ----------
    name: str
          Name of the logging object
    file_path: str
          Path of the file that stores the logs (will be created if not exists)
    level: default debug
          Logging level 

    Returns
    -------
    logger: obj
           Object that handles the logging through its inherent methods
    """
    levels = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "critical": logging.CRITICAL
    }

    logger = logging.getLogger(name)
    logger.setLevel(levels.get(level, logging.DEBUG))

    _init_logging_config(logger=logger, file_path=file_path)

    return logger


def _init_logging_config(logger: logging.Logger, file_path: str):
    """set the configuration of a logging object

    Set up the logging format, create the console and file handlers

    Parameters
    ----------
    logger:
           Logging object
    file_path: str
           Path of the file that stores the logs (will be created if not exists)    
    """
    console_format = logging.Formatter('[%(name)s][%(levelname)s] - %(message)s')
    file_format = logging.Formatter('[%(asctime)s][%(name)s][%(levelname)s] - %(message)s')

    console_handler = logging.StreamHandler(); console_handler.setLevel(logging.INFO); console_handler.setFormatter(console_format)
    file_handler = logging.FileHandler(file_path, mode="a"); file_handler.setLevel(logging.DEBUG); file_handler.setFormatter(file_format)

    logger.addHandler(console_handler); logger.addHandler(file_handler)



import io
import pathlib
import tarfile
from random import randint

from kubernetes import client, stream

from .utils import pod
from .utils.loggus import init_logger


logger = init_logger(__name__ if __name__ != "__main__" else pathlib.Path(__file__).stem, level="debug")


def _check_pvc_exists(api_instance: client.CoreV1Api, namespace: str, pvc_name: str):
     """Create a persistent volume claim (pvc) if it does not exist

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Kubernetes Namespace
     pvc_name: str
          Name used in the creation of the pvc
     Returns
     -------
     Bool
          Wether the PVC exists or not
     """
     try:
          pvcs = api_instance.list_namespaced_persistent_volume_claim(namespace=namespace)
     except Exception as e:
          logger.error("Could not list pvcs in namespace {n}. Erorr: {e}".format(n=namespace, e=e))
          raise e
     for pvc in pvcs.items:
          if pvc.metadata.name == pvc_name:
               return True
          
     return False


def _create_pvc(api_instance: client.CoreV1Api, namespace: str, pvc_name: str) -> None:
     """Create a persistent volume claim (pvc) if it does not exist

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Kubernetes Namespace
     pvc_name: str
          Name used in the creation of the pvc

     Returns
     -------
     None
          None
     """
     if _check_pvc_exists(api_instance, namespace, pvc_name):
         logger.info(f"PVC {pvc_name} already exists in namespace {namespace}. Skipping creation.")
         return
        
     # create pvc
     logger.info("Creating PVC {pvc} in namespace {namespace}.".format(pvc=pvc_name, namespace=namespace))

     body = client.V1PersistentVolume(
          api_version="v1",
          kind="PersistentVolumeClaim",
          metadata=client.V1ObjectMeta(name=pvc_name, namespace=namespace),
          spec=client.V1PersistentVolumeClaimSpec(
               access_modes=["ReadWriteMany"],
               resources=client.V1ResourceRequirements(requests={"storage": "100Gi"}), #TODO: make this configurable
               storage_class_name="rook-cephfs", #TODO: make this configurable
          )
     )
     api_instance.create_namespaced_persistent_volume_claim(namespace=namespace, body=body)

     logger.info("PVC {pvc} successfully created in namespace {namespace}".format(pvc=pvc_name, namespace=namespace))


def _delete_pvc(api_instance: client.CoreV1Api, namespace: str, pvc_name: str) -> None:
     """Delete a persistent volume claim (pvc) if it exists

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Kubernetes Namespace
     pvc_name: str
          Name used of the pvc to delete
     Returns
     -------
     None
          None
     """

     # check if pvc already exists
     if not _check_pvc_exists(api_instance, namespace, pvc_name):
          logger.info(f"PVC {pvc_name} not found in namespace {namespace}. Skipping deletion.")
          return
     
     try:
          _ = api_instance.delete_namespaced_persistent_volume_claim(
               name=pvc_name,
               namespace=namespace,
               body=client.V1DeleteOptions(
                    propagation_policy='Foreground',
               )
          )
          logger.info(f"PVC '{pvc_name}' deleted successfully.")
     except Exception as e:
          logger.error("Could not delete PVC {pvc_name}. Please delete manually! Error: {e}".format(pvc_name=pvc_name, e=e))
          

# taken from https://github.com/kubernetes-client/python/issues/476
# I don't know why there isn't a built-in function for this yet
def _copy_files_to_pod(api_instance: client.CoreV1Api, namespace: str, pod_name: str, source_path: pathlib.Path, dest_path: pathlib.Path) -> None:
     """Copy files to the temporary pod

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Namespace to use for creating pvc
     pod_name: str
          Name of the pod to copy files to
     source_path: pathlib.Path
          Path to the files to copy
     dest_path: pathlib.Path
          Mount path of the pvc in the pod

     Returns
     -------
     None
          None"""
     buf = io.BytesIO()
     with tarfile.open(fileobj=buf, mode='w:tar') as tar:  # To compress set 'w:gz'
          tar.add(source_path, arcname=dest_path.joinpath(source_path.name))
     buf.seek(0)

     exec_command = ['tar', 'xvf', '-', '-C', '/']   # To decompress set 'xzvf'
     resp = stream.stream(api_instance.connect_get_namespaced_pod_exec, pod_name, namespace,
                         command=exec_command,
                         stderr=True, stdin=True,
                         stdout=True, tty=False,
                         _preload_content=False)

     # copy data to pod in chunks to avoid issues with larger files
     chunk_size = 10 * 1024 * 1024
     while resp.is_open():
          resp.update(timeout=1)
          # if resp.peek_stdout():
          # logger.debug(f"STDOUT: {resp.read_stdout()}")
          if resp.peek_stderr():
               logger.error(f"STDERR: {resp.read_stderr()}")
               raise RuntimeError(f"Error copying files to pod: {resp.read_stderr()}")
          if read := buf.read(chunk_size):
               #logger.debug(f"Uploading chunk: {read}")
               logger.debug(f"Uploading chunkof size: {len(read)}")
               resp.write_stdin(read)
          else:
               break
     resp.close()

     logger.info("Files succesfully transfered to kube pvc!")


def _update_file_dest_name(api_instance: client.CoreV1Api, namespace: str, pod_name: str, old_name: pathlib.Path, new_name: pathlib.Path) -> None:
     """Copy files to the temporary pod

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Namespace to use for creating pvc
     pod_name: str
          Name of the pod to copy files to
     old_name: str
          Name of the directory where the files have been copied
     new_name: str
          Name of the new directory

     Returns
     -------
     None
          None"""

     exec_command = ["/bin/sh", "-c", 'cp -R ./{o}/* ./{n} && rm -R ./{o}/*'.format(o=old_name, n=new_name)]
     resp = stream.stream(api_instance.connect_get_namespaced_pod_exec, pod_name, namespace,
                         command=exec_command,
                         stderr=True, stdin=True,
                         stdout=True, tty=False,
                         _preload_content=False)

     while resp.is_open():
          resp.update(timeout=1)
          if resp.peek_stdout():
               print(f"STDOUT: \n{resp.read_stdout()}")
          if resp.peek_stderr():
               print(f"STDERR: \n{resp.read_stderr()}")

     resp.close()

     logger.info("Files directory successfully renamed to avoid conflicts!")


def upload_files_to_pvc(namespace: str, prefix: str, pvc: str, files_path: list[(pathlib.Path, pathlib.Path)]) -> str:
     """Upload model and data files to the pvc
     Parameters
     ----------
     kube: kubernetes client
          Kubernetes client
     pvc: str
          Name of pvc to use for data and model storage
     namespace: str
          Namespace to use for creating pvc
     Returns
     -------
     None
          None
     """
     api_instance = client.CoreV1Api()
     
     pvc_name = "{x}-{y}".format(x=prefix, y=pvc) if prefix else pvc
     pod_name = "{x}-seaquest-tmp-uploader".format(x=prefix)
     pod_name = pod.make_pod_name_unique(api_instance, namespace, pod_name)

     try:
          _create_pvc(api_instance, namespace, pvc_name)

          logger.info("Creating pod to move data to pvc ...")
          pod._launch_pod(api_instance, namespace, pod_name, pvc_name)

          pod._wait_for_running_state(api_instance, namespace, pod_name, 120) #TODO: make this configurable

          logger.info("Attempting to upload files to pvc ...")
          for (file_path, dest_dir) in files_path:
               dest_path = pathlib.PurePosixPath(pvc_name)
               if dest_dir is not None:
                   dest_path = dest_path.joinpath(dest_dir)
               _copy_files_to_pod(api_instance, namespace, pod_name, file_path, dest_path)

          # change dir name to avoid conflicts in case of multiple jobs with the same model and different config (e.g. different data files)
          # model_dir in _generate_runner_config should also be changed if this part changes

          new_name = pathlib.PurePosixPath(pvc_name).joinpath("_".join([files_path[1][1], files_path[1][0].stem]))
          _update_file_dest_name(api_instance, namespace, pod_name, dest_path, new_name)

     except Exception as e: # TODO: handle stuff here
          pod._delete_pod(api_instance, namespace, pod_name)
          logger.error(e)
          raise e

     pod._delete_pod(api_instance, namespace, pod_name)

     return pvc_name



import os
from pathlib import Path
from kubernetes import client, config

config.load_kube_config()
api_instance_core = client.CoreV1Api()
api_instance_batch = client.BatchV1Api()

namespace = "dl4nlpspace"
prefix = "iones"

pvc_name = "test-pvc"
pvc_name_prefix = "iones-test-pvc"

pod_name = "seaquest-test-pod"
pod_name_prefix = "iones-seaquest-test-pod"

pod_wait_timeout = 300 # seconds

config_file_path = Path(Path(__file__).parent.resolve()).joinpath("test_config_nogpu.yaml")
test_model_dir_path = Path(Path(__file__).parent.resolve()).joinpath("test_model_dir")
data_file_path = Path(Path(__file__).parent.resolve()).joinpath("test_model_dir").joinpath("test-weights.txt")

test_model_name = "ExampleModel"
COMPLETE_ARGS = ["-cf", str(config_file_path),
                  "-md", str(test_model_dir_path),
                  "-mn", test_model_name,
                  "-f", "infer",
                  "-p", "iones",
                  "-df", str(data_file_path)]

INCOMPLETE_ARGS_1= ["-md", "/models"]
INCOMPLETE_ARGS_2= ["-md", "/models", "-cf", str(config_file_path)]

test_job_name_prefix = "iones-seaquest-test-job"
job_output_dir = "iones-seaquest-test-job_output"
data_file = "test-weights"
job_arguments = {
    "-od": job_output_dir,
    "-md": "test_model_dir"
}

runner_arguments = [
    "-od", "output_dir",
    "-md", "test_model_dir",
    "-cf", "tests\\unit\\test_config_nogpu.yaml"
]
fun = "infer"

files_path = [(test_model_dir_path, None),
              (data_file_path, "test_model_dir")]

output_dir = "output_dir"




from random import randint
from time import sleep

from kubernetes import client

from .loggus import init_logger

logger = init_logger(__name__, level="debug")


def make_pod_name_unique(api_instance: client.CoreV1Api, namespace: str, pod_name: str) -> str:
     """Make pod name unique by appending a random number if a pod with the same name exists
     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Namespace to get pods from
     pod_name: str
          Desired pod name
     Returns
     -------
     str
          Unique pod name
     """
     all_pod_names = _get_list_of_pods(api_instance, namespace)
     pod_name = "{pod}-{rand}".format(pod=pod_name, rand=randint(1000, 9999))
     while pod_name in all_pod_names:    
          pod_name = "{pod}-{rand}".format(pod=pod_name, rand=randint(1000, 9999))

     return pod_name


def _get_list_of_pods(api_instance: client.CoreV1Api, namespace: str) -> list[str]:
     """Get list of pod names in a namespace

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Namespace to get pods from

     Returns
     -------
     list
          List of pod names
     """
     pods = api_instance.list_namespaced_pod(namespace=namespace)
     pod_names = [pod.metadata.name for pod in pods.items]
     return pod_names


def _launch_pod(api_instance: client.CoreV1Api, namespace: str, pod_name:str, pvc_name: str) -> None:
     """Launch a pod that will be quickly deleted after the operations are done

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Namespace to use for creating pod
     pod_name: str
          Name of pod
     pvc_name: str
          Name of pvc to bind the pod to

     Returns
     -------
     None
     """

     body = client.V1Pod(
          api_version="v1",
          kind="Pod",
          metadata=client.V1ObjectMeta(name=pod_name, namespace=namespace),
          spec=client.V1PodSpec(
               containers=[
                    client.V1Container(
                    name="alpine",
                    image="alpine:latest",
                    command=['tail', '-f', '/dev/null'],
                    volume_mounts=[client.V1VolumeMount(mount_path="/{x}".format(x=pvc_name), name=pvc_name)]
                    )
               ],
               volumes=[client.V1Volume(name=pvc_name, persistent_volume_claim=client.V1PersistentVolumeClaimVolumeSource(claim_name=pvc_name))],
          )
     )

     api_instance.create_namespaced_pod(namespace, body)

     logger.info("Succesfully created pod {pod}".format(pod=pod_name))


def _delete_pod(api_instance: client.CoreV1Api, namespace: str, pod_name: str) -> None:
     """Delete the temporary pod used to upload files to the pvc

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Namespace to use for creating pvc
     pod_name: str
          Name of the pod to delete

     Returns
     -------
     None
          None
     """
     all_pods = _get_list_of_pods(api_instance, namespace)
     if pod_name not in all_pods:
          logger.warning(f"Pod {pod_name} not found in namespace {namespace}. Skipping deletion.")
          return

     api_instance.delete_namespaced_pod(name=pod_name, namespace=namespace)

     logger.info("Successfully deleted pod {pod}.".format(pod=pod_name))


def _is_running(api_instance: client.CoreV1Api, pod_name: str, namespace: str, timeout: int) -> bool:
     """Check if the pod is done initializing and is running

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     pod_name: str
          Name of the pod to check
     namespace: str
          Namespace to use for creating pvc
     timeout: int
          Time to wait for pod to reach running state until raising error 

     Returns
     -------
     bool
          True if the pod is done initializing, False otherwise
     """

     # Watch and streaming solutions didn't work so we use this for now ): 
     # (this snippet is also present in the official examples)
     # decorators for timeout didn't work because function is unpicklable
     while True:
          resp = api_instance.read_namespaced_pod(name=pod_name, namespace=namespace)
          if resp.status.phase == 'Running':
               return True
          sleep(1)
          
          timeout -= 1
          if timeout == 0:
               raise TimeoutError("Timeout waiting for pod to be in running state.")


def _wait_for_running_state(api_instance: client.CoreV1Api, namespace: str, pod_name: str, timeout: int) -> bool:
     """Wait for the pod to be in running state, else timeout after 60 seconds

     Parameters
     ----------
     api_instance: kubernetes client
          Kubernetes client
     namespace: str
          Namespace to use for creating pvc
     pod_name: str
          Name of the pod to check
     timeout:
          Time to wait for pod to reach running state until raising error      
     
     Returns
     -------
     bool
          True if the pod is done initializing, else raise TimeoutError
     """
     try:
          _is_running(api_instance, pod_name, namespace, timeout)
     except TimeoutError:
          logger.error("Timeout error: Pod {pod} failed to start in {t} seconds".format(pod=pod_name, t=timeout))
          raise TimeoutError(f"Pod {pod_name} failed to start.")


from seaquest import runner
from seaquest.utils.validate import _parse_args, _parse_runner_args
from test_vars import *


def test_load_model_dir_pass():
    try:
        runner._load_model_dir(test_model_dir_path.name)
    except:
        assert False, "Model dir failed to load"


def test_load_model_dir_fail():
    try:
        runner._load_model_dir("nonexistent_path")
    except:
        pass


def test_load_model_class_pass():
    try:
        namespace = runner._load_model_dir(test_model_dir_path.name)
        runner._load_model_class(namespace, test_model_name)
    except:
        assert False, "Model class failed to load"


def test_load_model_class_fail():
    try:
        runner._load_model_class(None, test_model_name)
    except:
        pass


def test_main():
    args = _parse_args(COMPLETE_ARGS.copy())
    args_runner = _parse_runner_args(runner_arguments)

    args.update(args_runner)
    try:
        runner.main(args)
    except Exception as e:
        assert False, "Main function from runner failed to run: {e}".format(e=e)
    



import pathlib
import sys
import os

from kubernetes import client, config
from yaml import safe_dump

from seaquest import uploader, launcher
from seaquest.utils import validate, loggus, pod

config.load_kube_config()
logger = loggus.init_logger(__name__ if __name__ != "__main__" else pathlib.Path(__file__).stem)


def _generate_runner_config(args: dict) -> pathlib.Path:
    """generate and save the runner config file
    
    Parameters
    ----------
    args: dict
        Dictionary of parsed arguments
    Returns
    -------
    file_path: pathlib.Path
        Path to the location where the runner configuration is saved
    """

    try: 
        dir = pathlib.Path(args["md_dir"]).resolve()

        file_path = dir / "runner.yaml"
        args_cp = args.copy()
        args_cp["md_dir"] = "_".join([dir.name, pathlib.Path(args["data_file"]).stem])
        args_cp["data_file"] = pathlib.Path(args["data_file"]).name
        with open(file_path, "w") as file:
            safe_dump(args_cp, file)
    except Exception as e:
        logger.error("Could not generate runner config, aborting! Error: {e}".format(e=e))
        raise e

    return file_path


def _delete_runner_config(path: pathlib.Path) -> None:
    """remove the generated runner config file

    Parameters
    ----------
    path: pathlib.Path
        Path to the location where the runner configuration is saved
    Returns
    -------
    None
    """
    os.remove(path)


def _upload_files(args: dict) -> None:
    """upload the model directory and the data file to the PVC.

    Parameters
    ----------
    args: dict
        Dictionary of parsed arguments
    Returns
    -------
    None
    """
    pvc_name = args["pvc_params"]["pvc-name"] if "pvc_params" in args \
            else "seaquest-pvc"
    
    files = [(pathlib.Path(args["md_dir"]).resolve(), None), 
             (pathlib.Path(args["data_file"]).resolve(), pathlib.Path(args["md_dir"]).resolve().name)]
    try:
        pvc_name = uploader.upload_files_to_pvc(args["kube_env"]["namespace"], args["prefix"], pvc_name, \
                                    files)

    except Exception as e:
        logger.error("Could not upload the model and data to the pvc! Error: {e}".format(e=e))
        raise e
    
    return pvc_name


def _launch_jobs(args: dict, pvc_name: str) -> None:
    """launch the experimental jobs

    Parameters
    ----------
    args: dict
        Dictionary of parsed arguments
    pvc_name: str
        Name of PVC to mount
    Returns
    -------
    None
    """
    data_file = pathlib.Path(args["data_file"]).stem
    model_dir = "_".join([pathlib.Path(args["md_dir"]).name, data_file])

    api_instance = api_instance = client.BatchV1Api()
    try:
        all_created_jobs = launcher.create_jobs(api_instance, 1, args["kube_env"]["namespace"], args["job_spec"], args["prefix"], \
                            args["suffix"], args["model_name"], args["model_fun"], pvc_name, model_dir, data_file)
    except Exception as e:
        logger.error("Could not launch jobs!".format(e=e))
        raise e

    return all_created_jobs


def _save_config_and_jobs(args: dict) -> None:
    """save a file containing the names of the launched jobs to feed to the monitor module

    Parameters
    ----------
    args: dict
        Dictionary of parsed arguments
    Returns
    -------
    None
    """
    PATH_MAPPING = {
        "linux": "./config",
        "darwin": "Library/Preferences/seaquest",
        "win32": "AppData/Local/seaquest",
    }

    try:
        file_path = pathlib.Path(pathlib.Path.home()).joinpath(PATH_MAPPING[str(sys.platform).lower()]) 
        file_path.mkdir(parents=True, exist_ok=True)
        file_path = file_path / "seaquest.yaml" 
        
        with open(file_path, "w") as file:
            safe_dump(args, file)
    except Exception as e:
        logger.warning("Could not save monitor config file and created jobs names. Error: {e}".format(e=e))


def main(argv):
    """parse arguments, create a runner config, upload the model and data, launch the jobs save the names of the launched jobs.

    Parameters
    ----------
    argv: list
        List of arguments passed to the module
    Returns
    -------
    None
    """
    args = validate.parse_and_validate_args(argv)
    logger.info("Arguments parsed and validated succesfully!")

    # generate the runner config and save it in the model directory so we upload everything at once
    rc_path = _generate_runner_config(args)
    pvc_name = _upload_files(args)
    _delete_runner_config(rc_path)

    all_created_jobs = _launch_jobs(args, pvc_name)

    # save the output to a file: config used and jobs
    args["launched_jobs"] = all_created_jobs
    _save_config_and_jobs(args)


if __name__ == "__main__":
    main(sys.argv)


from seaquest import uploader
from seaquest.utils import pod
from test_vars import *


def test_setup_conditions():
    try:
        uploader._create_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        assert False, "Failed to setup testing environment"


def test_make_pod_name_unique():
    try:
        _ = pod.make_pod_name_unique(api_instance_core, namespace, pod_name_prefix)
    except Exception as e:
        assert False, "Could not generate a unique pod name!"


def test_launch_pod_pass():
    pod_name_tmp = None
    try:
        pod_name_tmp = pod.make_pod_name_unique(api_instance_core, namespace, pod_name_prefix)
        pod._launch_pod(api_instance_core, namespace, pod_name_tmp, pvc_name_prefix)

        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
    except Exception as e:    
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
        assert False, f"Launching temp pod failed with exception: {e}"


def test_launch_temp_pod_fail():
    pod_name_tmp = None
    try:
        # launching a pod with a non-existing pvc should fail
        pod_name_tmp = pod.make_pod_name_unique(api_instance_core, namespace, pod_name_prefix)
        pod._launch_pod(api_instance_core, namespace, pod_name_tmp, "no-pvc")
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
    except Exception as e:
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
        assert True, "Launching temp pod failed as expected since pvc does not exist."


def test_delete_pod():
    pod_name_tmp = None
    try:
        pod_name_tmp = pod.make_pod_name_unique(api_instance_core, namespace, pod_name_prefix)
        pod._launch_pod(api_instance_core, namespace, pod_name_tmp, pvc_name_prefix)
        
        pod._wait_for_running_state(api_instance_core, namespace, pod_name_tmp, pod_wait_timeout)
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
    except Exception as e:
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
        assert False, f"Deleting temp pod failed with exception: {e}"


def test_timeout_waiting_for_pod():
    pod_name_tmp = None
    try:
        pod_name_tmp = pod.make_pod_name_unique(api_instance_core, namespace, pod_name_prefix)
        pod._launch_pod(api_instance_core, namespace, pod_name_tmp, pvc_name_prefix)
        
        pod._wait_for_running_state(api_instance_core, namespace, pod_name_tmp, 1) # wait 1 second
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
    except TimeoutError:
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
        assert True
    except Exception as e:
        pod._delete_pod(api_instance_core, namespace, pod_name_tmp)
        assert False, f"Test failed with exception: {e}"


def test_clean_up():
    try:
        uploader._delete_pvc(api_instance_core, namespace, pvc_name_prefix)
    except Exception as e:
        assert False, "Cleanup failed with exception: {e}".format(e=e)


import os

from seaquest.utils import validate
from test_vars import *

def test_parse_args():
    args = validate._parse_args(COMPLETE_ARGS.copy())

    assert args["config_file"] == COMPLETE_ARGS[1]
    assert args["md_dir"] == COMPLETE_ARGS[3]
    assert args["model_name"] == COMPLETE_ARGS[5]
    assert args["model_fun"] == COMPLETE_ARGS[7]
    assert args["prefix"] == COMPLETE_ARGS[9]
    assert args["data_file"] == COMPLETE_ARGS[11]


def test_parse_args_fail():
    try:
        validate._parse_args(INCOMPLETE_ARGS_1.copy())
    except:
        assert True


def test_validate_complete_args():
    args = validate._parse_args(COMPLETE_ARGS.copy())
    assert validate._validate(args) is True


def test_validate_incomplete_args():
    args = validate._parse_args(INCOMPLETE_ARGS_2.copy())
    try:
        validate._validate(args)
    except Exception as e:
        assert True


import pathlib
import sys
import tarfile
from tempfile import TemporaryFile
from time import sleep
from yaml import safe_load

from kubernetes import client, config, stream

from .utils.loggus import init_logger
from .utils.validate import _parse_monitor_args
from .utils.pod import _launch_pod, _delete_pod, _wait_for_running_state, make_pod_name_unique

config.load_kube_config()
logger = init_logger(__name__ if __name__ != "__main__" else pathlib.Path(__file__).stem, level="debug")


def _delete_job(api_instance: client.BatchV1Api, namespace: str, job_name: str) -> None:
    """Delete a job
    
    Parameters
    ----------
    api_instance: kubernetes client
        Kubernetes client
    namespace: str
        Namespace name
    job_name: str
        The name of the job to be deleted
    Returns
    -------
    file_path: pathlib.Path
        Path to the location where the runner configuration is saved
    """

    api_instance.delete_namespaced_job(name=job_name, namespace=namespace, propagation_policy='Foreground')


def _delete_all_jobs(api_instance: client.BatchV1Api, namespace: str, job_names: str) -> None:
    """Delete all jobs in a list

    Parameters
    ----------
    api_instance: kubernetes client
        Kubernetes client
    namespace: str
        Namespace name
    job_names: list
        The names of the job to be deleted
    Returns
    -------
    None
    """

    for job in job_names:
        try:
            _delete_job(api_instance, namespace, job)
        except Exception as e:
            logger.error("Couldn't delete job {job}, please delete it manually! Error: {e}".format(job=job, e=e))


def _monitor_jobs(api_instance_batch: client.BatchV1Api, api_instance_core: client.CoreV1Api, namespace: str, job_names: list, pvc: str, dest_dir: pathlib.Path, prefix: str) -> None:
    """Monitor jobs until they reach a finish state (Succeeded or Failed)

    Parameters
    ----------
    api_instance_batch: kubernetes batch client
        Kubernetes batch client
    api_instance_core: kubernetes core client
        Kubernetes core client
    namespace: str
        Namespace name
    job_names: list
        The names of the job to be deleted
    pvc: str
        The name of the PVC
    dest_dir: str
        The name of the destination directory
    prefix: str
        The prefix
    Returns
    -------
    None
    """
    
    # Watch and streaming solutions didn't work so we use this for now ): 
    # (this snippet is also present in the official examples)
    logger.info("Started monitoring jobs!")
    logger.info("Monitored jobs: {jobs}".format(jobs=" ".join(job_names)))

    jobs_done = [False] * len(job_names)

    while True:
        for idx, job in enumerate(job_names):

            if jobs_done[idx]:
                continue
            
            resp = None
            try:
                resp = api_instance_batch.read_namespaced_job_status(name=job, namespace=namespace)
            except Exception as e:
                jobs_done[idx] = True
                logger.info("Monitoring for job {job} has failed with error: {err}".format(job=job, err=e))
                continue

            if resp.status.succeeded:
                try:
                    _pull_files(api_instance_core, namespace, prefix, job, pvc, dest_dir)
                except Exception as e:
                    logger.error("Couldn't pull files for job {job}. Skipping. Error: {e}".format(job=job, e=e))
                stat = "Succeeded"

            elif resp.status.failed:
                stat = "Failed"

            elif not resp.status.active or resp.status.active > 0:
                continue # still initiating or running
            
            jobs_done[idx] = True
            logger.info("Job {job} has finished with status `{stat}`".format(job=job, stat=stat))
            _delete_job(api_instance_batch, namespace, job)

            sleep(1)

        if all(jobs_done) == True:
            logger.info("All jobs are done! Results from successful jobs have been pulled to {dir}".format(dir=str(dest_dir)))
            return
        
        sleep(60)


def _pull_files(api_instance: client.CoreV1Api, namespace:str, prefix: str, job_name:str, pvc: str, dest_dir: str) -> None:
    """Pull results to a local directory

    Parameters
    ----------
    api_instance_core: kubernetes core client
        Kubernetes core client
    namespace: str
        Namespace name
    prefix: str
        The prefix
    job_name: str
        The name of the job to be deleted
    pvc: str
        The name of the PVC
    dest_dir: str
        The name of the destination directory
    Returns
    -------
    None
    """
    logger.info("Pulling output of job {job} ...".format(job=job_name))

    pvc_name = "{x}-{y}".format(x=prefix, y=pvc) if prefix else pvc
    pod_name = "{x}-seaquest-tmp-uploader".format(x=prefix)
    pod_name = make_pod_name_unique(api_instance, namespace, pod_name)

    _launch_pod(api_instance, namespace, pod_name, pvc_name)
    _wait_for_running_state(api_instance, namespace, pod_name, timeout=120) # TODO: make this configurable

    files_path = pathlib.Path("{p}-{pvc}".format(p=prefix, pvc=pvc)).joinpath("{job}_output".format(job=job_name))

    try:
        _copy_files_from_pod(api_instance, pod_name, namespace, files_path, dest_dir)
        _delete_pod(api_instance, namespace, pod_name)
    except Exception as e:
        _delete_pod(api_instance, namespace, pod_name)
        logger.error("Copy files for job {job} has failed with error {err}".format(job=job_name, err=e))
        raise e


# https://github.com/prafull01/Kubernetes-Utilities/blob/master/kubectl_cp_as_python_client.py
# They should really make a cp method already ):
def _copy_files_from_pod(api_instance: client.CoreV1Api, pod_name: str, namespace: str, src_path: pathlib.Path, dest_dir: pathlib.Path) -> None:
    """Copy files to a local directory through a pod

    Parameters
    ----------
    api_instance_core: kubernetes core client
        Kubernetes core client
    pod_name: str
        The name of the pod
    namespace: str
        Namespace name
    src_path: pathlib.Path
        Path on the PVC mount where to copy the files from
    dest_dir: pathlib.Path
        The name of the local destination directory where to copy the files to 
    Returns
    -------
    None
    """

    logger.info("Starting to copy files from kube storage through pod {pod}".format(pod=pod_name))
    exec_command = ['tar', 'czf', '-', src_path.as_posix()] # for some reason cd-ing and then tar-ing does not return any std output

    with TemporaryFile() as tar_buffer:
        resp = stream.stream(api_instance.connect_get_namespaced_pod_exec, pod_name, namespace,
                      command=exec_command,
                      binary=True,
                      stderr=True, stdin=True,
                      stdout=True, tty=False,
                      _preload_content=False)

        while resp.is_open():
            resp.update(timeout=1)
            if resp.peek_stdout():
                out = resp.read_stdout()
                logger.info("Got {num} bytes".format(num=len(out)))
                tar_buffer.write(out)
            if resp.peek_stderr():
                logger.error("STDERR: %s" % resp.read_stderr().decode('utf-8'))
        resp.close()

        tar_buffer.flush()
        tar_buffer.seek(0)

        with tarfile.open(fileobj=tar_buffer, mode='r:gz') as tar:
            subdir_and_files = [
                tarinfo for tarinfo in tar.getmembers()
            ]
            tar.extractall(path=dest_dir, members=subdir_and_files)

    logger.info("Files pulled to {dest}".format(dest=str(dest_dir)))


def _load_config():
    """Load monitor config

    Parameters
    ----------
    None
    Returns
    -------
    None
    """
    PATH_MAPPING = {
        "linux": "./config",
        "darwin": "Library/Preferences/seaquest",
        "win32": "AppData/Local/seaquest",
    }

    file_path = pathlib.Path(pathlib.Path.home()).joinpath(PATH_MAPPING[str(sys.platform).lower()]) 
    file_path.mkdir(parents=True, exist_ok=True)
    file_path = file_path / "seaquest.yaml" 

    with open(file_path, "r") as file:
        args = safe_load(file)

    return args


def main(argv):
    # check system arguments
    passed_args = _parse_monitor_args(argv)
    dest_dir = pathlib.Path(passed_args["dest_dir"]).resolve()

    try:
        # load config file from default location if not specified
        if "config_file" not in passed_args:
            saved_args = _load_config()
        else:
            with open(passed_args["config_file"], "r") as file:
                saved_args = safe_load(file)
    except Exception as e:
        logger.error("Could not load config file! Error {e}".format(e=e))
        raise e

    # if config file found run start monitoring jobs
    api_instance_batch = client.BatchV1Api()
    api_instance_core = client.CoreV1Api()

    _monitor_jobs(api_instance_batch, api_instance_core, saved_args["kube_env"]["namespace"], saved_args["launched_jobs"], saved_args["pvc_params"]["pvc-name"], dest_dir, saved_args["prefix"])


if __name__ == "__main__":
    main(sys.argv)